// var sclient = require('socket.io-client');
var EventEmitter = require('events').EventEmitter;
var util = require('util');
var utils = require('./util/utils');
var protocol = require('./util/protocol');
var logger = require('pomelo-logger').getLogger('pomelo-admin', __filename);
var net = require('net');
var Composer = require('@nd/simple-stream-pkg').Composer

var ST_INITED = 1;
var ST_CONNECTED = 2;
var ST_REGISTERED = 3;
var ST_CLOSED = 4;
var STATUS_INTERVAL = 30 * 1000; //
var RECONNECT_INTERVAL = 10 * 1000;
var RECONNECT_MAX_FAILED = 20;

/**
 * MonitorAgent Constructor
 *
 * @class MasterAgent
 * @constructor
 * @param {Object} opts construct parameter
 *                      opts.consoleService {Object} consoleService
 *                      opts.id {String} server id
 *                      opts.type {String} server type, 'master', 'connector', etc.
 *                      opts.info {Object} more server info for current server, {id, serverType, host, port}
 * @api public
 */
var MonitorAgent = function (opts) {
  EventEmitter.call(this);
  this.consoleService = opts.consoleService;
  this.id = opts.id;
  this.type = opts.type;
  this.info = opts.info;
  this.socket = null;
  this.reqId = 1;
  this.callbacks = {};
  this.state = ST_INITED;
  this.retry = 0;
};

util.inherits(MonitorAgent, EventEmitter);

module.exports = MonitorAgent;

/**
 * register and connect to master server
 *
 * @param {String} port
 * @param {String} host
 * @param {Function} cb callback function
 * @api public
 */
MonitorAgent.prototype.connect = function (port, host, cb) {
  if (this.state > ST_INITED) {
    logger.error('monitor client has connected or closed.');
    return;
  }
  function pack(composer, event, body) {
    return composer.compose(JSON.stringify({ event, body }));
  }
  let socket = net.createConnection(port, host);
  this.socket = socket;
  socket.composer = new Composer();
  socket.pack = pack.bind(null, socket.composer);
  socket.recvEvt = new EventEmitter();

  socket.composer.on('error', (err) => {
    logger.error('socket composer error -- %j', err);
    socket.end();
  });

  socket.composer.on('data', (data) => {
    const pkg = JSON.parse(data.toString());
    socket.recvEvt.emit(pkg.event, pkg.body);
  });


  socket.on('data', (d) => {
    socket.composer.feed(d);
  });

  // sclient.connect('http://' + host + ':' + port, { 'force new connection': true, 'reconnect': true, 'max reconnection attempts': 20 });

  var self = this;
  socket.recvEvt.on('register', function (msg) {
    if (msg && msg.code === protocol.PRO_OK) {
      self.state = ST_REGISTERED;
      utils.invokeCallback(cb);

      setTimeout(() => {
        const req = {
          id: self.id,
          type: 'monitor',
          serverType: self.type,
          pid: process.pid,
          seq: 0,
        };


          self.socket.write(self.socket.pack('heartbeat', req));

      }, STATUS_INTERVAL);

    } else {
      self.emit('close');
      logger.error('server %j %j register master failed', self.id, self.type);
    }
  });

  socket.recvEvt.on('monitor', function (msg) {
    if (self.state !== ST_REGISTERED) {
      return;
    }

    msg = protocol.parse(msg);

    if (msg.command) {
      // a command from master
      self.consoleService.command(msg.command, msg.moduleId, msg.body, function (err, res) {
        //notify should not have a callback
      });
    } else {
      if (msg.respId) {
        // a response from monitor
        var cb = self.callbacks[msg.respId];
        if (!cb) {
          logger.warn('unknown resp id:' + msg.respId);
          return;
        }
        delete self.callbacks[msg.respId];
        utils.invokeCallback(cb, msg.error, msg.body);
        return;
      }

      // request from master
      self.consoleService.execute(msg.moduleId, 'monitorHandler', msg.body, function (err, res) {
        if (protocol.isRequest(msg)) {
          var resp = protocol.composeResponse(msg, err, res);
          if (resp) {
            self.socket.write(self.socket.pack('monitor', resp));
          }
        } else {
          //notify should not have a callback
          logger.error('notify should not have a callback.');
        }
      });
    }
  });

  socket.on('connect', function () {
    if (self.state > ST_INITED) {
      //ignore reconnect
      return;
    }
    self.retry = 0;
    self.state = ST_CONNECTED;
    var req = {
      id: self.id,
      type: 'monitor',
      serverType: self.type,
      pid: process.pid,
      info: self.info
    };
    var authServer = self.consoleService.authServer;
    var env = self.consoleService.env;
    authServer(req, env, function (token) {
      req['token'] = token;
      self.socket.write(self.socket.pack('register', req));
    });
  });

  socket.on('error', function (err) {
    if (self.state < ST_CONNECTED) {
      if (self.retry > RECONNECT_MAX_FAILED) {
        // logger.error('connect failed, reconnect', self.retry);
      // error occurs during connecting stage
      utils.invokeCallback(cb, err);
    } else {
        setTimeout(() => {
          logger.info('connect failed, reconnecting', host, port);
          self.connect(port, host, cb);
        }, RECONNECT_INTERVAL);
        self.retry = self.retry + 1;
        logger.error('connect failed, reconnect', self.retry);
      }

    } else {
      self.emit('error', err);
    }
  });

  socket.on('end', function (reason) {
    logger.warn(reason, this.id, this.type, this.info);
    self.state = ST_CLOSED;
    self.emit('close');
  });

  socket.recvEvt.on('reconnect', function () {
    self.state = ST_CONNECTED;
    var req = {
      id: self.id,
      type: 'monitor',
      serverType: self.type,
      pid: process.pid,
      info: self.info
    };
    self.socket.write(self.socket.pack('reconnecting', req));
  });

  socket.recvEvt.on('reconnect_ok', function (msg) {
    if (msg && msg.code === protocol.PRO_OK) {
      self.state = ST_REGISTERED;
    }
  });

  socket.recvEvt.on('heartbeat', function (msg) {

    const req = {
      id: self.id,
      type: 'monitor',
      serverType: self.type,
      pid: process.pid,
      seq: msg.seq + 1,
    };

    setTimeout(() => {
      self.socket.write(self.socket.pack('heartbeat', req));
    }, STATUS_INTERVAL);

  });
};

/**
 * close monitor agent
 *
 * @api public
 */
MonitorAgent.prototype.close = function () {
  if (this.state >= ST_CLOSED) {
    return;
  }
  this.state = ST_CLOSED;
  this.socket.end();
};

/**
 * set module
 *
 * @param {String} moduleId module id/name
 * @param {Object} value module object
 * @api public
 */
MonitorAgent.prototype.set = function (moduleId, value) {
  this.consoleService.set(moduleId, value);
};

/**
 * get module
 *
 * @param {String} moduleId module id/name
 * @api public
 */
MonitorAgent.prototype.get = function (moduleId) {
  return this.consoleService.get(moduleId);
};

/**
 * notify master server without callback
 *
 * @param {String} moduleId module id/name
 * @param {Object} msg message
 * @api public
 */
MonitorAgent.prototype.notify = function (moduleId, msg) {
  if (this.state !== ST_REGISTERED) {
    logger.error('agent can not notify now, state:' + this.state);
    return;
  }
  this.socket.write(this.socket.pack('monitor', protocol.composeRequest(null, moduleId, msg)));
};

MonitorAgent.prototype.request = function (moduleId, msg, cb) {
  if (this.state !== ST_REGISTERED) {
    logger.error('agent can not request now, state:' + this.state);
    return;
  }
  var reqId = this.reqId++;
  this.callbacks[reqId] = cb;
  this.socket.write(this.socket.pack('monitor', protocol.composeRequest(reqId, moduleId, msg)));
};
